#include <_9syscall.h>
#include <errno9.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stdint.h>
#include <sys/uio.h>
#include <stdbool.h>
#include <string.h>
#include <poll.h>
#include <stdlib.h>
#include <time.h>
#include <sys/select.h>
#include <_9.h>
#include "lock.h"
#include "fork.h"

#define BUF_LEN 1024

typedef struct buffer
{
	char buf[BUF_LEN];
	size_t start;
	size_t end;
} buffer;

typedef struct fd_info
{
	long lock;
	int err;

	buffer reader;
	buffer writer;

	int flags;

	bool buffered;
	bool open;
} fd_info;

size_t _fdinfo_len = 1024;
fd_info _fdinfo_start[1024];
fd_info *_fdinfo = _fdinfo_start;

void _fdinit(void)
{
	for (int i = 0; i <= 3; ++i)
	{
		_fdinfo[i].open = true;
		_fdinfo[i].lock = UNLOCKED;
		_fdinfo[i].flags = O_RDWR;
	}
}

static inline void ensure(size_t fd)
{
	if (fd < _fdinfo_len) return;

	size_t old_len = _fdinfo_len;
	fd_info *old_fdinfo = _fdinfo;

	_fdinfo_len = fd * 1.5;
	_fdinfo = malloc(_fdinfo_len * sizeof(_fdinfo[0]));
	memcpy(_fdinfo, old_fdinfo, old_len * sizeof(_fdinfo[0]));
	memset(_fdinfo + old_len, 0, (_fdinfo_len - old_len) * sizeof(_fdinfo[0]));
}

static ssize_t nonblock_read_partial(int fd, void *buf, size_t count)
{
	int sem = TRYACQUIRE(_fdinfo[fd].lock);

	if (sem == 0) return -EAGAIN;
	else if (sem == -1) return -EIO;

	if (_fdinfo[fd].err)
	{
		RELEASE(_fdinfo[fd].lock);
		return _fdinfo[fd].err;
	}

	size_t len = _fdinfo[fd].reader.end - _fdinfo[fd].reader.start;

	if (count <= len)
	{
		memcpy(buf, &_fdinfo[fd].reader.buf[_fdinfo[fd].reader.start], count);
		_fdinfo[fd].reader.start += count;
		RELEASE(_fdinfo[fd].lock);
		return count;
	}

	/* Copy whatever we already have read */
	memcpy(buf, &_fdinfo[fd].reader.buf[_fdinfo[fd].reader.start], len);
	count -= len;

	int fork = _9syscall_1(SYS_RFORK, RFPROC | RFNOWAIT | RFMEM);
	switch (fork)
	{
	case -1:
		_9syscall_2(SYS_SEMRELEASE, (uvlong) &_fdinfo[fd].lock, 1);
		return -EIO;
	case 0:
		if (count > BUF_LEN) count = BUF_LEN;
		_fdinfo[fd].reader.start = 0;
		_fdinfo[fd].reader.end = 0;
		int ret = _9syscall_4(
				SYS_PREAD,
				fd,
				(uvlong) _fdinfo[fd].reader.buf,
				count,
				-1
			);

		if (ret < 0) _fdinfo[fd].err = _syserrno();
		if (ret == 0) _fdinfo[fd].err = EAGAIN;
		else _fdinfo[fd].reader.end = count;
		RELEASE(_fdinfo[fd].lock);
		return -EIO;
	default: return len;
	}
}

static ssize_t nonblock_write_partial(int fd, const void *buf, size_t count)
{
	int sem = TRYACQUIRE(_fdinfo[fd].lock);

	if (sem == 0) return -EAGAIN;
	else if (sem == -1) return -EIO;

	if (_fdinfo[fd].err)
	{
		RELEASE(_fdinfo[fd].lock);
		return _fdinfo[fd].err;
	}

	if (count > BUF_LEN) count = BUF_LEN;
	
	memcpy(_fdinfo[fd].writer.buf, buf, count);
	_fdinfo[fd].writer.start = 0;
	_fdinfo[fd].writer.end = count;

	/* Flush reader */
	_fdinfo[fd].reader.start = 0;
	_fdinfo[fd].reader.end = 0;

	int fork = _9syscall_1(SYS_RFORK, RFPROC | RFNOWAIT | RFMEM);
	switch (fork)
	{
	case -1:
		RELEASE(_fdinfo[fd].lock);
		return -EIO;
	case 0:
		do
		{
			size_t len = _fdinfo[fd].writer.end - _fdinfo[fd].writer.start;
			size_t res = len - _9syscall_4(
					SYS_PWRITE,
					fd,
					(uvlong) _fdinfo[fd].writer.buf[_fdinfo[fd].writer.start],
					len,
					-1
				);
			if (res < 0)
			{
				_fdinfo[fd].err = _syserrno();
				RELEASE(_fdinfo[fd].lock);
				return -EIO;
			}
			_fdinfo[fd].writer.start += res;
		} while ( _fdinfo[fd].writer.start < _fdinfo[fd].writer.end);

		_fdinfo[fd].writer.start = 0;
		_fdinfo[fd].writer.end = 0;
		RELEASE(_fdinfo[fd].lock);
		return -EIO;
	default: return count;
	}
}

static ssize_t buffered_read(int fd, void *buf, size_t count)
{
	size_t total = count;
	while (count)
	{
		ssize_t r = nonblock_read_partial(fd, buf, count);
		if (r < 0) return r;
		count -= r;
		buf += r;
	}
	return total;
}

static ssize_t buffered_write(int fd, const void *buf, size_t count)
{
	size_t total = count;
	while (count)
	{
		ssize_t r = nonblock_write_partial(fd, buf, count);
		if (r < 0) return r;
		count -= r;
		buf += r;
	}
	return total;
}

static ssize_t nonblock_read(int fd, void *buf, size_t count)
{
	ssize_t r;
	size_t total = 0;

	do {
		r = nonblock_read_partial(fd, buf + total, count - total);
		total += r;
	} while (total < count && r > 0);

	return r < 0 ? r : total;
}

static ssize_t nonblock_write(int fd, const void *buf, size_t count)
{
	size_t total = 0;
	ssize_t r = nonblock_write_partial(fd, buf, count);
	if (r == 0) r = nonblock_write_partial(fd, buf, count);
	if (r < 0) return r;
	total += r;

	while (total < count && r > 0)
	{
		r = nonblock_write_partial(fd, buf + total, count - total);
		total += r;
	}
	return r < 0 ? r : total;
}

ssize_t _9write(int fd, const void *buf, size_t count)
{
	if (!_fdinfo[fd].open || _fdinfo[fd].flags & O_RDONLY) return -EBADF;
	if (count == 0) return 0;
	if (_fdinfo[fd].flags & O_NONBLOCK) return nonblock_write(fd, buf, count);
	else if (_fdinfo[fd].buffered) return buffered_write(fd, buf, count);

	ssize_t ret = _9syscall_4(SYS_PWRITE, fd, (uintptr_t) buf, count, -1LL);
	return ret < 0 ? -_syserrno() : ret;
}

ssize_t _9read(int fd, void *buf, size_t count)
{
	if (!_fdinfo[fd].open || _fdinfo[fd].flags & O_WRONLY) return -EBADF;
	if (count == 0) return 0;
	if (_fdinfo[fd].flags & O_NONBLOCK) return nonblock_read(fd, buf, count);
	else if (_fdinfo[fd].buffered) return buffered_read(fd, buf, count);

	ssize_t ret = _9syscall_4(SYS_PREAD, fd, (uintptr_t) buf, count, -1LL);
	return ret < 0 ? -_syserrno() : ret;
}

ssize_t _9pwrite(int fd, const void *buf, size_t count, off_t offset)
{
	/* TODO: Handle buffered fds */
	if (!_fdinfo[fd].open || _fdinfo[fd].flags & O_RDONLY) return -EBADF;
	if (count == 0) return 0;
	ssize_t ret = _9syscall_4(SYS_PWRITE, fd, (uintptr_t) buf, count, offset);
	return ret < 0 ? -_syserrno() : ret;
}

ssize_t _9pread(int fd, void *buf, size_t count, off_t offset)
{
	/* TODO: Handle buffered fds */
	if (!_fdinfo[fd].open || _fdinfo[fd].flags & O_WRONLY) return -EBADF;
	if (count == 0) return 0;
	ssize_t ret = _9syscall_4(SYS_PREAD, fd, (uintptr_t) buf, count, offset);
	return ret < 0 ? -_syserrno() : ret;
}

int _9open(const char *file, int flags, int mode)
{
	int omode = 0;
	int ret;

	if (strcmp(file, "/dev/urandom") == 0) file = "/dev/random";

	if ((flags & O_ACCMODE) == O_RDONLY) omode |= OREAD;
	else if ((flags & O_ACCMODE) == O_WRONLY) omode |= OWRITE;
	else omode |= ORDWR;

	if (flags & O_TRUNC) omode |= OTRUNC;
	if (flags & O_CLOEXEC) omode |= OCEXEC;

	if (flags & O_CREAT)
		ret = _9syscall_3(SYS_CREATE, (uvlong) file, omode, mode & 0777);
	else ret = _9syscall_2(SYS_OPEN, (uvlong) file, omode);

	if (ret >= 0)
	{
		ensure(ret);
		_fdinfo[ret].flags = flags;
		_fdinfo[ret].lock = UNLOCKED;
		_fdinfo[ret].reader.start = 0;
		_fdinfo[ret].reader.end = 0;
		_fdinfo[ret].open = true;
	}

	int err = _syserrno();
	return ret < 0 ? -err : ret;
}

int _9creat(const char *file, int mode)
{
	int ret = _9syscall_3(
			SYS_CREATE,
			(uvlong) file,
			O_WRONLY | O_CREAT | O_TRUNC,
			mode & 0777
		);
	int err = _syserrno();
	return ret < 0 ? -err : ret;
}

int _9close(int fd)
{
	if (TRYACQUIRE(_fdinfo[fd].lock) == -1)
		return -EIO;

	/* Flush reader */
	_fdinfo[fd].reader.start = 0;
	_fdinfo[fd].reader.end = 0;
	_fdinfo[fd].open = false;

	ssize_t ret = _9syscall_1(SYS_CLOSE, fd);
	int err = _syserrno();

	RELEASE(_fdinfo[fd].lock);
	return err ? -err : ret;
}

ssize_t _9writev(int fd, const struct iovec *iov, int iovcnt)
{
	size_t sum = 0;
	int e_old;
	for (int i = 0; i < iovcnt; ++i)
	{
		size_t n = _9write(fd, iov[i].iov_base, iov[i].iov_len);
		int err = _syserrno();
		e_old = err != 0 ? err : e_old;
		if (n != iov[i].iov_len) return -1;
		sum += n;
	}
	return e_old ? -e_old : sum;
}

ssize_t _9readv(int fd, const struct iovec *iov, int iovcnt) {
	ssize_t sum = 0;
	int e_old = 0;
	for (int i = 0; i < iovcnt; ++i) {
		ssize_t n = _9read(fd, iov[i].iov_base, iov[i].iov_len);
		int err = _syserrno();
		if (err)
		{
			e_old = err;
			break;
		}
		if (n < 0) return -1;
		sum += n;
		if (n < iov[i].iov_len) break;
	}
	return e_old ? -e_old : sum;
}

ssize_t _9lseek(int fd, off_t offset, int whence)
{
	int64_t ret;
	int err = 0;
	if (TRYACQUIRE(_fdinfo[fd].lock) == -1)
		return -EIO;

	/* Flush reader */
	_fdinfo[fd].reader.start = 0;
	_fdinfo[fd].reader.end = 0;

	if (_9syscall_4(SYS_SEEK, (uvlong) &ret, fd, offset, whence) < 0)
		err = _syserrno();

	RELEASE(_fdinfo[fd].lock);
	return err ? -err : ret;
}

static bool isreadable(int fd)
{
	if (!_fdinfo[fd].buffered && !(_fdinfo[fd].flags & O_NONBLOCK))
	{
		_fdinfo[fd].buffered = true;
		_fdinfo[fd].reader.start = 0;
		_fdinfo[fd].reader.end = 0;
		return isreadable(fd);
	}

	if (TRYACQUIRE(_fdinfo[fd].lock) <= 0) return false;
	RELEASE(_fdinfo[fd].lock);

	if (_fdinfo[fd].reader.end - _fdinfo[fd].reader.start > 0) return true;

	nonblock_read_partial(fd, NULL, 1); /* Initialize read */
	return false;
}

static bool iswritable(int fd)
{
	if (!_fdinfo[fd].buffered && !(_fdinfo[fd].flags & O_NONBLOCK))
	{
		_fdinfo[fd].buffered = true;
		_fdinfo[fd].reader.start = 0;
		_fdinfo[fd].reader.end = 0;
		return true;
	}

	return TRYACQUIRE(_fdinfo[fd].lock);
}

uint64_t __nsec(void);
int _9poll(struct pollfd *fds, nfds_t nfds, int timeout)
{
	uint64_t start = __nsec();
	int count = 0;

	if (timeout == 0) return 0;

	for (nfds_t i = 0; i < nfds; ++i)
	{
		fds[i].revents = 0;
		if (fds[i].events == 0) fds[i].events = POLLNVAL | POLLHUP | POLLERR;
		if (fds[i].events & POLLRDNORM) fds[i].events |= POLLIN;
		if (fds[i].events & POLLWRNORM) fds[i].events |= POLLOUT;
	}

	while (count == 0 && (timeout < 0 || __nsec() - start < timeout * 1000000))
	{
		for (nfds_t i = 0; i < nfds; ++i)
		{
			if (fds[i].fd < 0) continue;
			if (fds[i].events & POLLNVAL && !_fdinfo[fds[i].fd].open)
			{
				fds[i].revents |= POLLNVAL;
				++count;
				continue;
			}

			if (fds[i].events & POLLIN && isreadable(fds[i].fd))
			{
				fds[i].revents |= POLLIN;
				++count;
			}

			if (fds[i].events & POLLOUT && iswritable(fds[i].fd))
			{
				fds[i].revents |= POLLOUT;
				++count;
			}

			if ((fds[i].events & (POLLHUP | POLLERR)) && _fdinfo[fds[i].fd].err)
			{
				fds[i].revents |= POLLHUP | POLLERR;
				++count;
			}
		}
	}
	return count;
}

int _9select(
		int nfds,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout)
{
	int i;
	struct pollfd pfds[1024] = {0};

	for (i = 0; i < nfds; ++i)
	{
		if (FD_ISSET(i, readfds))
		{
			pfds[i].fd = i;
			pfds[i].events |= POLLIN;
		}
		if (FD_ISSET(i, writefds))
		{
			pfds[i].fd = i;
			pfds[i].events |= POLLOUT;
		}
		if (FD_ISSET(i, exceptfds))
		{
			pfds[i].fd = i;
			pfds[i].events |= POLLERR;
		}
	}

	int ret;
	if (timeout == NULL) ret = _9poll(pfds, i, -1);
	else
	{
		long msec = timeout->tv_sec * 1000 + timeout->tv_usec / 1000;
		ret = poll(pfds, i, msec);
	}

	FD_ZERO(readfds);
	FD_ZERO(writefds);
	FD_ZERO(exceptfds);

	for (i = 0; i < nfds; ++i)
	{
		if (pfds[i].revents & POLLIN) FD_SET(pfds[i].fd, readfds);
		if (pfds[i].revents & POLLOUT) FD_SET(pfds[i].fd, writefds);
		if (pfds[i].revents & POLLERR) FD_SET(pfds[i].fd, exceptfds);
	}
	return ret;
}

static int dup_sys(int oldfd, int newfd)
{
	int new = _9syscall_2(SYS_DUP, oldfd, newfd);
	if (new < 0) return -EBADF;
	ensure(new);
	_fdinfo[new] = _fdinfo[oldfd];
	return new;
}
int _9dup2(int oldfd, int newfd)
{
	if (!_fdinfo[oldfd].open || newfd < 0) return -EBADF;
	return dup_sys(oldfd, newfd);
}

int _9dup(int oldfd)
{
	if (!_fdinfo[oldfd].open) return -EBADF;
	return dup_sys(oldfd, -1);
}

int _9fcntl(int fd, int cmd, long arg)
{
	if (!_fdinfo[fd].open) return -EBADF;
	switch (cmd)
	{
	case F_GETFD: return 0;
	case F_GETFL: return _fdinfo[fd].flags;
	case F_SETFD:
	case F_SETFL:
		/* TODO: O_APPEND, O_ASYNC, O_DIRECT, O_NOATIME */
		_fdinfo[fd].flags = (arg & ~O_ACCMODE) |
			(_fdinfo[fd].flags & O_ACCMODE);
		return 0;
	case F_DUPFD:
	case F_DUPFD_CLOEXEC:
	{
		while (arg < _fdinfo_len && _fdinfo[fd].open) ++arg;
		size_t cap = arg + _fdinfo_len + 1024;
		int newfd = -1;
		while (newfd < 0 && newfd < cap) newfd = _9dup2(fd, arg++);
		if (newfd < 0) return -_syserrno();
		return newfd;
	}
	default:
		_FAIL_PRINT("cross9: fcntl - cmd %d not implemented\n", cmd);
		return -ENOSYS;
	}
}

int _9unlink(const char *pathname)
{
	/* TODO: Handle open files, apparently APE does more than this for them */
	if (_9syscall_1(SYS_REMOVE, (uvlong) pathname) < 0) return -_syserrno();
	return 0;
}

ssize_t _9readlink(
		const char *restrict pathname,
		char *restrict buf,
		size_t bufsiz)
{
	buf[0] = 0;
	strncpy(buf, pathname, bufsiz);
	return strnlen(buf, bufsiz);
}
