#include <signal.h>
#include <stddef.h>
#include <string.h>
#include <_9syscall.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <_9.h>
#define _GNU_SOURCE /* gettid */
#include <unistd.h>
#include "lock.h"

#define NCONT 0
#define NDFLT 1
#define NSAVE 2
#define NRSTR 3

static void mask_unset(unsigned long *p, int n)
{
	int idx = n / (sizeof(unsigned long) * 8);
	int bit = n % (sizeof(unsigned long) * 8);
	p[idx] &= ~(1LU << bit);
}

static void mask_set(unsigned long *p, int n)
{
	int idx = n / (sizeof(unsigned long) * 8);
	int bit = n % (sizeof(unsigned long) * 8);
	p[idx] |= (1LU << bit);
}

static bool mask_get(const unsigned long *p, int n)
{
	int idx = n / (sizeof(unsigned long) * 8);
	int bit = n % (sizeof(unsigned long) * 8);
	return p[idx] >> bit;
}

/* sigtab is taken from APE */
static struct {
	char    *msg;     /* just check prefix */
	int    num;
} sigtab[] = {
	{"hangup",                          SIGHUP},
	{"interrupt",                       SIGINT},
	{"quit",                            SIGQUIT},
	{"alarm",                           SIGALRM},
	{"sys: trap: illegal instruction",  SIGILL},
	{"sys: trap: reserved instruction", SIGILL},
	{"sys: trap: reserved",             SIGILL},
	{"sys: trap: arithmetic overflow",  SIGFPE},
	{"abort",                           SIGABRT},
	{"sys: fp:",                        SIGFPE},
	{"exit",                            SIGKILL},
	{"die",                             SIGKILL},
	{"kill",                            SIGKILL},
	{"sys: trap: bus error",            SIGSEGV},
	{"sys: trap: address error",        SIGSEGV},
	{"sys: trap: TLB",                  SIGSEGV},
	{"sys: write on closed pipe",       SIGPIPE},
	{"term",                            SIGTERM},
	{"usr1",                            SIGUSR1},
	{"usr2",                            SIGUSR2},
	{"virtual alarm",                   SIGVTALRM},
	{"profiling alarm",                 SIGPROF},
};
#define NSIGTAB ((sizeof sigtab)/(sizeof (sigtab[0])))

static struct sigaction sighdlr[_NSIG];

typedef struct sigset_list
{
	int tid;
	sigset_t set;
	struct sigset_list *next;
} sigset_list;
static sigset_list procset_first;
static sigset_list *procset; /* List of {thread, sigset} */

/* Linked list */
typedef struct sig_list
{
	int signum;
	struct sig_list *next;
} sig_list;
static sig_list *pending;

static sig_list *lock_queue;
static long sig_lock; /* This should probably be thread-local */
static long lock_queue_lock = 1;

__attribute__((noreturn))
static void handle(int n);

void _9siglock(void)
{
	++sig_lock;
}

void _9sigunlock(void)
{
	if (sig_lock == 0) return;
	ACQUIRE(lock_queue_lock);
	if (--sig_lock == 0 && lock_queue)
	{
		sig_list *cur = lock_queue;
		while (cur)
		{
			sig_list *next = cur->next;
			handle(cur->signum);
			free(cur);
			cur = next;
		}
		lock_queue = NULL;
	}
	RELEASE(lock_queue_lock);
}

extern void (*notehandler)();
void _siginit(void)
{
	sigset_list *old = procset;
	procset = &procset_first;
	procset->next = old;
	procset->tid = gettid();
	memset(&procset->set, 0, sizeof(procset->set));
	_9syscall_1(SYS_NOTIFY, (uvlong) &notehandler);
}

static sigset_t *gettset(void)
{
	sigset_list *pair;
	int tid = gettid();
	for (pair = procset; pair; pair = pair->next)
		if (pair->tid == tid) break;
	return pair ? &pair->set : NULL;
}

__attribute__((noreturn))
void _9sigrestore_rt(void)
{
	_9syscall_1(SYS_NOTED, NCONT);
	__builtin_unreachable();
}

__attribute__((noreturn))
static void ndefault(void)
{
	_9syscall_1(SYS_NOTED, NDFLT);
	__builtin_unreachable();
}

__attribute__((noreturn))
static void handle(int n)
{
	sigset_t *ts = gettset();
	if (!ts)
	{
		puts("Failed to get thread set");
		fflush(stdout);
		ndefault();
	}

	void (*f)(int) = sighdlr[sigtab[n].num].sa_handler;
	if (sig_lock)
	{
		ACQUIRE(lock_queue_lock);
		sig_list *new = malloc(sizeof(sig_list));
		new->next = NULL;
		new->signum = sigtab[n].num;
		
		if (lock_queue == NULL) lock_queue = new;
		else
		{
			sig_list *cur = lock_queue;
			while (cur->next) cur = cur->next;
			cur->next = new;
		}
		f = SIG_IGN;
		RELEASE(lock_queue_lock);
	}
	else if (mask_get(ts->__bits, sigtab[n].num) && f != SIG_IGN)
	{
		sig_list *new = malloc(sizeof(sig_list));
		new->next = NULL;
		new->signum = sigtab[n].num;

		if (pending == NULL) pending = new;
		else
		{
			sig_list *cur = pending;
			while (cur->next) cur = cur->next;
			cur->next = new;
		}
		f = SIG_IGN;
	}

	if (f == SIG_IGN)
	{
		_9sigrestore_rt();
		__builtin_unreachable();
	}

	if (f == SIG_DFL || f == SIG_ERR)
	{
		ndefault();
		__builtin_unreachable();
	}

	f(sigtab[n].num);
	sighdlr[sigtab[n].num].sa_restorer();
	__builtin_unreachable();
}

void _notehandler(char *msg)
{
	int i;
	void (*f)(int);

#ifdef SIGTRACE
	printf("note: %s\n", msg);
	fflush(stdout);
#endif
	
	for (i = 0; i < NSIGTAB; ++i)
		if (strncmp(msg, sigtab[i].msg, strlen(sigtab[i].msg)) == 0)
			handle(sigtab[i].num);
	ndefault();
	__builtin_unreachable();
}

int _9rt_sigaction(
		int signum,
		const struct sigaction *restrict act,
		struct sigaction *restrict oact,
		size_t sigsetsize)
{
	if (signum < 1 || signum >= _NSIG ||
			signum == SIGKILL || signum == SIGSTOP ||
			sigsetsize > sizeof(sigset_t))
		return -EINVAL;

	if (oact) *oact = sighdlr[signum];

	if (act == NULL) return 0;

	struct sigaction newact = *act;
	if (newact.sa_flags & SA_SIGINFO)
	{
		printf("rt_sigaction: Unimplemented flag - SA_SIGINFO\n");
		return ENOSYS;
	}

	if (!(newact.sa_flags & SA_RESTORER)) newact.sa_restorer = _9sigrestore_rt;

	sighdlr[signum] = newact;
	return 0;
}

int _9rt_sigprocmask(
		int how,
		const sigset_t *set,
		sigset_t *oldset,
		size_t sigsetsize)
{
	if (sigsetsize > sizeof(sigset_t)) return -EINVAL;

	if (oldset) *oldset = procset->set;

	sigset_t *ts = gettset();
	if (ts == NULL) return -EINVAL;

	if (set)
	{
		switch (how)
		{
		case SIG_BLOCK:
			for (size_t i = 0; i < _NSIG; ++i)
				if (mask_get(set->__bits, i)) mask_set(ts->__bits, i);
			break;
		case SIG_UNBLOCK:
			for (size_t i = 0; i < _NSIG; ++i)
				if (mask_get(set->__bits, i)) mask_unset(ts->__bits, i);
			break;
		case SIG_SETMASK:
			*ts = *set;
			break;
		default: return -EINVAL;
		}
	}

	sig_list *cur = pending;
	while (cur)
	{
		if (!mask_get(ts->__bits, cur->signum))
		{
			sig_list *prev = cur;
			sig_list *next = cur->next;
			handle(cur->signum);
			free(cur);
			prev->next = next;
			cur = next;
		} else cur = cur->next;
	}
	return 0;
}
